﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Input;

namespace SolSoft.DataBinding.WpfDemo
{
	public class MainViewModel : BaseViewModel
	{
		public MainViewModel()
		{
			
			this.m_value1Property = this.CreateNotifyProperty(() => Value1);
			this.m_value2Property = this.CreateNotifyProperty(() => Value2);
			this.m_value3Property = this.CreateNotifyProperty(() => Value3);
			this.m_value4Property = this.CreateNotifyProperty(() => Value4);
			this.m_totalValueProperty = this.CreateDerivedNotifyProperty(() => TotalValue,
				m_value1Property, m_value2Property, m_value3Property, m_value4Property,
				CalculateTotal);

			this.m_isBusyProperty = IsBusyNotifyProperty.Create(this, () => IsBusy);
			this.m_isEnabledProperty = this.CreateDerivedNotifyProperty(() => IsEnabled, m_isBusyProperty, isBusy => !isBusy);

			this.m_generateValuesCommand = new DelegateCommandFromProperty(m_isEnabledProperty, GenerateValues);
			this.m_clearValuesCommand = new DelegateCommandFromProperty(m_isEnabledProperty, ClearValues);

			this.m_quitCommand = new DelegateCommand(() => true, RequestClose);
		}

		#region Values
		private readonly NotifyProperty<int?> m_value1Property;
		public int? Value1
		{
			get
			{
				return m_value1Property.Value;
			}
			set
			{
				m_value1Property.SetValue(value);

			}
		}

		private readonly NotifyProperty<int?> m_value2Property;
		public int? Value2
		{
			get
			{
				return m_value2Property.Value;
			}
			set
			{
				m_value2Property.SetValue(value);

			}
		}

		private readonly NotifyProperty<int?> m_value3Property;
		public int? Value3
		{
			get
			{
				return m_value3Property.Value;
			}
			set
			{
				m_value3Property.SetValue(value);

			}
		}

		private readonly NotifyProperty<int?> m_value4Property;
		public int? Value4
		{
			get
			{
				return m_value4Property.Value;
			}
			set
			{
				m_value4Property.SetValue(value);

			}
		}

		#endregion

		#region Total

		private readonly DerivedNotifyProperty<int?> m_totalValueProperty;
	
		/// <summary>
		/// Calculates the total, if any, from the given values
		/// </summary>
		private int? CalculateTotal(int? value1, int? value2, int? value3, int? value4)
		{
			int? toReturn = null;

			if (value1 != null)
				toReturn = (toReturn ?? 0) + value1;

			if (value2 != null)
				toReturn = (toReturn ?? 0) + value2;

			if (value3 != null)
				toReturn = (toReturn ?? 0) + value3;

			if (value4 != null)
				toReturn = (toReturn ?? 0) + value4;

			return toReturn;
		}
		public int? TotalValue
		{
			get
			{
				return m_totalValueProperty.Value;
			}
		}
		#endregion

		#region Busy/Enabled
		private readonly IsBusyNotifyProperty m_isBusyProperty;
		/// <summary>
		/// Determines whether the UI is busy
		/// </summary>
		public bool IsBusy
		{
			get
			{
				return m_isBusyProperty.Value;
			}
			protected set
			{
				m_isBusyProperty.SetValue(value);
			}
		}

		private readonly DerivedNotifyProperty<bool> m_isEnabledProperty;
		/// <summary>
		/// Determines whether the UI is enabled
		/// </summary>
		public bool IsEnabled
		{
			get
			{
				return m_isEnabledProperty.Value;
			}
		}
		#endregion

		/// <summary>
		/// Generates a random number for each value, with a delay in between to simulate processing
		/// </summary>
		private async void GenerateValues()
		{
			Random r = new Random();

			IsBusy = true;
			try
			{
				var delay = TimeSpan.FromSeconds(.66);
				const int min = 5;
				const int max = 15;

				await Task.Delay(delay);
				Value1 = r.Next(min, max);

				await Task.Delay(delay);
				Value2 = r.Next(min, max);

				await Task.Delay(delay);
				Value3 = r.Next(min, max);

				await Task.Delay(delay);
				Value4 = r.Next(min, max);
			}
			finally
			{
				IsBusy = false;
			}
		}

		private readonly DelegateCommand m_generateValuesCommand;
		public ICommand GenerateValuesCommand
		{
			get
			{
				return m_generateValuesCommand;
			}
		}

		/// <summary>
		/// Clears out the values by sequentially counting them down to 0
		/// </summary>
		private async void ClearValues()
		{
			IsBusy = true;
			try
			{
				var delay = TimeSpan.FromSeconds(.2);

				while (Value1 > 0)
				{
					await Task.Delay(delay);
					Value1--;
				}
				Value1 = null;

				while (Value2 > 0)
				{
					await Task.Delay(delay);
					Value2--;
				}
				Value2 = null;

				while (Value3 > 0)
				{
					await Task.Delay(delay);
					Value3--;
				}
				Value3 = null;

				while (Value4 > 0)
				{
					await Task.Delay(delay);
					Value4--;
				}
				Value4 = null;

			}
			finally
			{
				IsBusy = false;
			}
		}

		private readonly DelegateCommand m_clearValuesCommand;
		public ICommand ClearValuesCommand
		{
			get
			{
				return m_clearValuesCommand;
			}
		}


		public event EventHandler<EventArgs> CloseRequested;
		/// <summary>
		/// Requests that the application be closed
		/// </summary>
		public void RequestClose()
		{
			OnCloseRequested(EventArgs.Empty);
		}
		protected virtual void OnCloseRequested(EventArgs e)
		{
			CloseRequested.Raise(this, e);
		}
		private readonly DelegateCommand m_quitCommand;

		public ICommand QuitCommand
		{
			get
			{
				return m_quitCommand;
			}
		}
	}
}